@@ -33,17 +33,6 @@
 #include <regex.h>
 #include <stdlib.h>
 
-/* openssl deprecated API emulation */
-#ifndef HAVE_EVP_MD_CTX_NEW
-#define EVP_MD_CTX_new()	EVP_MD_CTX_create()
-#endif
-#ifndef HAVE_EVP_MD_CTX_FREE
-#define EVP_MD_CTX_free(ctx)	EVP_MD_CTX_destroy((ctx))
-#endif
-#ifndef HAVE_EVP_MD_CTX_RESET
-#define EVP_MD_CTX_reset(ctx)	EVP_MD_CTX_cleanup((ctx))
-#endif
-
 #ifdef ENABLE_NLS
 #include <libintl.h>
 #include <locale.h>
@@ -507,7 +496,8 @@ err:
 
 static int key_find(pam_handle_t *pamh, int flags, const char *user,
 		PKCS11_CTX *ctx, PKCS11_SLOT *slots, unsigned int nslots,
-		PKCS11_SLOT **authslot, PKCS11_KEY **authkey)
+		PKCS11_SLOT **authslot, PKCS11_KEY **authkey,
+		EVP_PKEY **authpubkey, PKCS11_CERT **authcert)
 {
 	int token_found = 0;
 
@@ -517,6 +507,7 @@ static int key_find(pam_handle_t *pamh, int flags, const char *user,
 
 	*authkey = NULL;
 	*authslot = NULL;
+	*authcert = NULL;
 
 	/* search all valuable slots for a key that is authorized by the user */
 	while (0 < nslots) {
@@ -532,6 +523,14 @@ static int key_find(pam_handle_t *pamh, int flags, const char *user,
 			break;
 		}
 		token_found = 1;
+		/* Update "slots" pointer: PKCS11 slots are implemented as array,
+		 * so starting to look at slot + 1 and decrementing nslots accordingly
+		 * will search the rest of slots. */
+		nslots -= (slot + 1 - slots);
+		slots = slot + 1;
+
+		if (slot->token->initialized == 0)
+			continue;
 
 		if (slot->token->loginRequired && slot->token->userPinLocked) {
 			pam_syslog(pamh, LOG_DEBUG, "%s: PIN locked",
@@ -551,10 +550,8 @@ static int key_find(pam_handle_t *pamh, int flags, const char *user,
 				if (1 != r) {
 					r = match_user_openssh(pubkey, user);
 				}
-				if (NULL != pubkey) {
-					EVP_PKEY_free(pubkey);
-				}
 				if (1 == r) {
+					*authpubkey = pubkey;
 					*authkey = keys;
 					*authslot = slot;
 					pam_syslog(pamh, LOG_DEBUG, "Found %s",
@@ -577,14 +574,9 @@ static int key_find(pam_handle_t *pamh, int flags, const char *user,
 				if (1 != r) {
 					r = match_user_openssh(pubkey, user);
 				}
-				if (NULL != pubkey) {
-					EVP_PKEY_free(pubkey);
-				}
 				if (1 == r) {
-					*authkey = PKCS11_find_key(certs);
-					if (NULL == *authkey) {
-						continue;
-					}
+					*authpubkey = pubkey;
+					*authcert = certs;
 					*authslot = slot;
 					pam_syslog(pamh, LOG_DEBUG, "Found %s",
 							certs->label);
@@ -596,20 +588,15 @@ static int key_find(pam_handle_t *pamh, int flags, const char *user,
 				count--;
 			}
 		}
-
-		/* Try the next possible slot: PKCS11 slots are implemented as array,
-		 * so starting to look at slot++ and decrementing nslots accordingly
-		 * will search the rest of slots. */
-		slot++;
-		nslots -= (slot - slots);
-		slots = slot;
-		pam_syslog(pamh, LOG_DEBUG, "No authorized key found");
+		pam_syslog(pamh, LOG_DEBUG, "No authorized key found on token %s",
+				slot->token->label);
 	}
 
 	if (0 == token_found) {
 		prompt(flags, pamh, PAM_ERROR_MSG , NULL, _("No token found"));
 	} else {
-		prompt(flags, pamh, PAM_ERROR_MSG , NULL, _("No authorized keys on token"));
+		prompt(flags, pamh, PAM_ERROR_MSG, NULL,
+				_("Could not find authorized keys on any of the tokens."));
 	}
 
 	return 0;
@@ -631,7 +618,7 @@ static int randomize(pam_handle_t *pamh, unsigned char *r, unsigned int r_len)
 	return ok;
 }
 
-static int key_verify(pam_handle_t *pamh, int flags, PKCS11_KEY *authkey)
+static int key_verify(pam_handle_t *pamh, int flags, PKCS11_KEY *authkey, EVP_PKEY *pubkey)
 {
 	int ok = 0;
 	unsigned char challenge[30];
@@ -640,7 +627,6 @@ static int key_verify(pam_handle_t *pamh, int flags, PKCS11_KEY *authkey)
 	const EVP_MD *md = EVP_sha1();
 	EVP_MD_CTX *md_ctx = EVP_MD_CTX_new();
 	EVP_PKEY *privkey = PKCS11_get_private_key(authkey);
-	EVP_PKEY *pubkey = PKCS11_get_public_key(authkey);
 
 	if (NULL == privkey)
 		goto err;
@@ -695,6 +681,8 @@ PAM_EXTERN int pam_sm_authenticate(pam_handle_t * pamh, int flags, int argc,
 	PKCS11_CTX *ctx;
 	unsigned int nslots;
 	PKCS11_KEY *authkey;
+	PKCS11_CERT *authcert;
+	EVP_PKEY *authpubkey = NULL;
 	PKCS11_SLOT *slots, *authslot;
 	const char *user;
 	const char *pin_regex;
@@ -706,12 +694,23 @@ PAM_EXTERN int pam_sm_authenticate(pam_handle_t * pamh, int flags, int argc,
 	}
 
 	if (1 != key_find(pamh, flags, user, ctx, slots, nslots,
-				&authslot, &authkey)) {
+				&authslot, &authkey, &authpubkey, &authcert)) {
 		r = PAM_AUTHINFO_UNAVAIL;
 		goto err;
 	}
-	if (1 != key_login(pamh, flags, authslot, pin_regex)
-			|| 1 != key_verify(pamh, flags, authkey)) {
+
+	if (1 != key_login(pamh, flags, authslot, pin_regex)) {
+		r = PAM_CRED_INSUFFICIENT;
+		goto err;
+	}
+
+	if (authkey == NULL && authcert) {
+		if (NULL == (authkey = PKCS11_find_key(authcert))) {
+			r = PAM_AUTHINFO_UNAVAIL;
+			goto err;
+		}
+	}
+	if (1 != key_verify(pamh, flags, authkey, authpubkey)) {
 		if (authslot->token->userPinLocked) {
 			r = PAM_MAXTRIES;
 		} else {
@@ -768,6 +767,8 @@ PAM_EXTERN int pam_sm_chauthtok(pam_handle_t * pamh, int flags, int argc,
 	PKCS11_CTX *ctx;
 	unsigned int nslots;
 	PKCS11_KEY *authkey;
+	PKCS11_CERT *authcert;
+	EVP_PKEY *authpubkey = NULL;
 	PKCS11_SLOT *slots, *authslot;
 	const char *user, *pin_regex;
 
@@ -785,7 +786,7 @@ PAM_EXTERN int pam_sm_chauthtok(pam_handle_t * pamh, int flags, int argc,
 	}
 
 	if (1 != key_find(pamh, flags, user, ctx, slots, nslots,
-				&authslot, &authkey)) {
+				&authslot, &authkey, &authpubkey, &authcert)) {
 		r = PAM_AUTHINFO_UNAVAIL;
 		goto err;
 	}
@@ -809,6 +810,7 @@ PAM_EXTERN int pam_sm_chauthtok(pam_handle_t * pamh, int flags, int argc,
 	r = PAM_SUCCESS;
 
 err:
+	EVP_PKEY_free(authpubkey);
 #ifdef TEST
 	module_data_cleanup(pamh, global_module_data, r);
 #endif
